@@ -19,7 +19,8 @@
 #'     brms_crossedRE <- brm(mpg ~ wt + (1|cyl) + (1+wt|gear), data = mtcars,
 #'            iter = 500, chains = 2)
 #'  }
-#'  if (require("brms")) {
+#'  \donttest{
+#'    ## too slow for CRAN (>5 seconds)
 #'    ## load stored object
 #'    load(system.file("extdata", "brms_example.rda", package="broom.mixed"))
 #'
@@ -29,6 +30,9 @@
 #'    tidy(fit, effects = "fixed", conf.method="HPDinterval")
 #'    tidy(fit, effects = "ran_vals")
 #'    tidy(fit, effects = "ran_pars", robust = TRUE)
+#'    if (require("posterior")) {
+#'      tidy(fit, effects = "ran_pars", rhat = TRUE, ess = TRUE)
+#'    }
 #'    # glance method
 #'    glance(fit)
 #'    ## this example will give a warning that it should be run with
@@ -55,19 +59,23 @@ NULL
 #'   than mean and standard deviation, to derive point estimates and uncertainty
 #' @param conf.int If \code{TRUE} columns for the lower (\code{conf.low})
 #' and upper bounds (\code{conf.high}) of posterior uncertainty intervals are included.
+#' @param exponentiate  whether to exponentiate the fixed-effect coefficient estimates and confidence intervals (common for logistic regression); if \code{TRUE}, also scales the standard errors by the exponentiated coefficient, transforming them to the new scale
 #' @param conf.level Defines the range of the posterior uncertainty conf.int,
 #'  such that \code{100 * conf.level}\% of the parameter's posterior distributio
 #'  lies within the corresponding interval.
 #'  Only used if \code{conf.int = TRUE}.
 #' @param conf.method method for computing confidence intervals
 #' ("quantile" or "HPDinterval")
+#' @param rhat whether to calculate the *Rhat* convergence metric
+#' (\code{FALSE} by default)
+#' @param ess whether to calculate the *effective sample size* (ESS) convergence metric
+#' (\code{FALSE} by default)
 #' @param fix.intercept rename "Intercept" parameter to "(Intercept)", to match
 #' behaviour of other model types?
 #' @param looic Should the LOO Information Criterion (and related info) be
-#'   included? See \code{\link[rstanarm]{loo.stanreg}} for details. Note: for
-#'   models fit to very large datasets this can be a slow computation.
+#'   included? See \code{\link[rstan]{loo.stanfit}} for details. (This
+#'   can be slow for models fit to large datasets.)
 #' @param ... Extra arguments, not used
-#'
 #' @return
 #' When \code{parameters = NA}, the \code{effects} argument is used
 #' to determine which parameters to summarize.
@@ -97,14 +105,30 @@ NULL
 #' in previous versions of the package), while technically inappropriate in
 #' a Bayesian setting where "fixed" and "random" effects are not well-defined,
 #' are used for compatibility with other (frequentist) mixed model types.
+#' @note At present, the components of parameter estimates are separated by parsing the column names of \code{as_draws} (e.g. \code{r_patient[1,Intercept]} for the random effect on the intercept for patient 1, or \code{b_Trt1} for the fixed effect \code{Trt1}. We try to detect underscores in parameter names and warn, but detection may be imperfect.
 #' @export
 tidy.brmsfit <- function(x, parameters = NA,
                          effects = c("fixed", "ran_pars"),
-                         robust = FALSE, conf.int = TRUE,
-                         conf.level = 0.95,
+                         robust = FALSE,
+                         conf.int = TRUE, conf.level = 0.95,
                          conf.method = c("quantile", "HPDinterval"),
+                         rhat = FALSE, ess = FALSE,
                          fix.intercept = TRUE,
+                         exponentiate = FALSE,
                          ...) {
+
+  check_dots(...)
+
+  std.error <- NULL ## NSE/code check
+  if (!requireNamespace("brms", quietly=TRUE)) {
+      stop("can't tidy brms objects without brms installed")
+  }
+  xr <- brms::restructure(x)
+  has_ranef <- nrow(xr$ranef)>0
+  if (any(grepl("_", rownames(fixef(x)))) ||
+        (has_ranef && any(grepl("_", names(ranef(x)))))) {
+      warning("some parameter names contain underscores: term naming may be unreliable!")
+  }
   use_effects <- anyNA(parameters)
   conf.method <- match.arg(conf.method)
   is.multiresp <- length(x$formula$forms)>1
@@ -115,32 +139,33 @@ tidy.brmsfit <- function(x, parameters = NA,
       sprintf("%s(%s)", pref, paste(unlist(x), collapse = "|"))
   }
   ## NOT USED:  could use this (or something like) to
-  ##  obviate need for gsub("_","",str_extract(...)) pattern ...  
+  ##  obviate need for gsub("_","",str_extract(...)) pattern ...
   prefs_LB <- list(
       fixed = "b_", ran_vals = "r_",
       ## don't want to remove these pieces, so use look*behind*
       ran_pars =   sprintf("(?<=(%s))", c("sd_", "cor_", "sigma")),
       components = sprintf("(?<=%s)", c("zi_","disp_"))
-    )
-    prefs <- list(
+  )
+  prefs <- list(
       fixed = "b_", ran_vals = "r_",
       ## no lookahead (doesn't work with grep[l])
       ran_pars = c("sd_", "cor_", "sigma"),
       components = c("zi_", "disp_")
-    )
-    pref_RE <- mkRE(prefs[effects])
+  )
+  pref_RE <- mkRE(prefs[effects])
   if (use_effects) {
     ## prefixes distinguishing fixed, random effects
 
     parameters <- pref_RE
   }
-  samples <- brms::posterior_samples(x, parameters)
-  if (is.null(samples)) {
+  samples_perchain <- brms::as_draws_array(x, parameters, regex = TRUE)
+  if (is.null(samples_perchain) || posterior::nvariables(samples_perchain) == 0) {
     stop("No parameter name matches the specified pattern.",
       call. = FALSE
     )
   }
-  terms <- names(samples)
+  samples <- brms::as_draws_matrix(samples_perchain)
+  terms <- colnames(samples)
   if (use_effects) {
       if (is.multiresp) {
         if ("ran_pars" %in% effects && any(grepl("^sd",terms))) {
@@ -155,7 +180,7 @@ tidy.brmsfit <- function(x, parameters = NA,
         ## t0 <- lapply(resp0,
         ##          function(x) if (length(x)==2) x[1] else x[-(length(x)-1)])
         ## t1 <- lapply(t0,
-        ##          function(x)     
+        ##          function(x)
         ##              case_when(
         ##                  x[[1]]=="b"  ~ sprintf("b%s",x[[2]]),
         ##                  x[[2]]=="sd" ~ sprintf("sd_%s__%s",x[[2]],x[[3]]),
@@ -205,25 +230,33 @@ tidy.brmsfit <- function(x, parameters = NA,
         }
       }
       res_list$ran_pars <-
-        data_frame(
+        dplyr::tibble(
           group = sapply(ss2, grpfun),
           term = sapply(ss2, termfun)
         )
     }
     if ("ran_vals" %in% effects) {
       rterms <- grep(mkRE(prefs$ran_vals), terms, value = TRUE)
-      ss <- strsplit(rterms, "(_+|\\[|\\]|,)")
-      termfun <- function(x) x[[length(x)]]
-      grpfun <- function(x) x[[2]]
-      levelfun <- function(x) x[[length(x)-1]]
+
+      vals <- stringr::str_match_all(rterms, "_(.+?)\\[(.+?),(.+?)\\]")
+
       res_list$ran_vals <-
-        data_frame(
-          group = sapply(ss, grpfun),
-          term = sapply(ss, termfun),
-          level = sapply(ss, levelfun)
+        dplyr::tibble(
+          group = purrr::map_chr(vals, function (v) { v[[2]] }),
+          term = purrr::map_chr(vals, function (v) { v[[4]] }),
+          level = purrr::map_chr(vals, function (v) { v[[3]] })
         )
+
     }
     out <- dplyr::bind_rows(res_list, .id = "effect")
+    # In the case where nrow(res_list$fixed) > 0 but nrow(res_list$ran_pars) == 0,
+    # the out object needs to be fixed a bit (replace columns with unexpected
+    # lists of NULL by expected vectors of NA).
+    for (col in c("group", "term")) {
+      if (is.list(out[[col]]) && all(sapply(out[[col]], is.null))) {
+        out[[col]] <- rep(NA, nrow(out))
+      }
+    }
     v <- if (fixed.only) seq(nrow(out)) else is.na(out$term)
     newterms <- stringr::str_remove(terms[v], mkRE(prefs[c("fixed")]))
     if (fixed.only) {
@@ -237,22 +270,37 @@ tidy.brmsfit <- function(x, parameters = NA,
     ## prefixes already removed for ran_vals; don't remove for ran_pars
   } else {
     ## if !use_effects
-    out <- data_frame(term = names(samples))
+    out <- dplyr::tibble(term = terms)
   }
   pointfun <- if (robust) stats::median else base::mean
   stdfun <- if (robust) stats::mad else stats::sd
   out$estimate <- apply(samples, 2, pointfun)
   out$std.error <- apply(samples, 2, stdfun)
   if (conf.int) {
+
     stopifnot(length(conf.level) == 1L)
     probs <- c((1 - conf.level) / 2, 1 - (1 - conf.level) / 2)
     if (conf.method == "HPDinterval") {
-      out[, c("conf.low", "conf.high")] <-
-          coda::HPDinterval(coda::as.mcmc(samples), prob=conf.level)
+        cc <- coda::HPDinterval(coda::as.mcmc(samples), prob=conf.level)
     } else {
-      out[, c("conf.low", "conf.high")] <-
-        t(apply(samples, 2, stats::quantile, probs = probs))
+        cc <- t(apply(samples, 2, stats::quantile, probs = probs))
     }
+    out$conf.low <- cc[,1]
+    out$conf.high <- cc[,2]
+  }
+  posterior_metrics <- c()
+  if (rhat) {
+    posterior_metrics <- c(posterior_metrics, rhat = posterior::rhat)
+  }
+  if (ess) {
+    posterior_metrics <- c(posterior_metrics, ess = posterior::ess_basic)
+  }
+  if (length(posterior_metrics) > 0) {
+    if (!requireNamespace("posterior", quietly=TRUE)) {
+        stop(paste0(paste0(names(posterior_metrics), collapse=", "),
+             " calculation for brmsfit objects requires posterior package"))
+    }
+    out[names(posterior_metrics)] <- posterior::summarise_draws(samples_perchain, posterior_metrics)[names(posterior_metrics)]
   }
   ## figure out component
   out$component <- dplyr::case_when(grepl("(^|_)zi",out$term) ~ "zi",
@@ -260,6 +308,14 @@ tidy.brmsfit <- function(x, parameters = NA,
                                     grepl("^disp",out$term) ~ "disp",
                                     TRUE ~ "cond")
 
+  if (exponentiate) {
+    vv <- c("estimate", "conf.low", "conf.high")
+    out <- (out
+      %>% mutate(across(contains(vv), exp))
+      %>% mutate(across(std.error, ~ . * estimate))
+    )
+  }
+
   out$term <- stringr::str_remove(out$term,mkRE(prefs[["components"]],
                                                 LB=TRUE))
   if (fix.intercept) {
@@ -274,8 +330,16 @@ tidy.brmsfit <- function(x, parameters = NA,
   return(out)
 }
 
-#' @rdname brms_tidiers
 
+#' @importFrom stats quantile
+#' @export
+sigma.brmsfit <- function (object, ...)  {
+    if (!("sigma" %in% brms::variables(object)))
+        return(1)
+    stats::quantile(brms::as_draws_array(object, "sigma"), probs=0.5)
+}
+
+#' @rdname brms_tidiers
 #' @export
 glance.brmsfit <- function(x, looic = FALSE, ...) {
   ## defined in rstanarm_tidiers.R
@@ -299,10 +363,13 @@ augment.brmsfit <- function(x, data = stats::model.frame(x), newdata = NULL,
   ## allow optional arguments to augment, e.g. pred.type,
   ## residual.type, re.form ...
   pred <- do.call(stats::predict, args)
-  ret <- dplyr::data_frame(.fitted = pred[, "Estimate"])
+  ret <- dplyr::tibble(.fitted = pred[, "Estimate"])
   if (se.fit) ret[[".se.fit"]] <- pred[, "Est.Error"]
   if (is.null(newdata)) {
     ret[[".resid"]] <- stats::residuals(x)[, "Estimate"]
+    ret <- dplyr::bind_cols(as_tibble(data), ret)
+  } else {
+    ret <- dplyr::bind_cols(as_tibble(newdata), ret)
   }
   return(ret)
 }
